package tools.descartes.prisma.agent.wildfly.callstack.impl;

import java.util.Collection;
import java.util.concurrent.ConcurrentHashMap;

import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InterfaceRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationGroupRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.TraceVisitor;
import tools.descartes.prisma.agent.wildfly.service.ModelExtractionService;

public class DefaultInterfaceRecord extends InterfaceRecord {
	
	private final ConcurrentHashMap<String, OperationRecord> operations = new ConcurrentHashMap<>();

	public DefaultInterfaceRecord(ComponentRecord component, String interf) {
		super(component, interf);
	}
	
	public OperationRecord requireOperation(String operation) {
		OperationRecord record = operations.get(operation);
		if (record == null) {
			record = getTraceFactory().createOperationRecord(this, operation);
			if (operations.putIfAbsent(operation, record) != null) {
				// a different thread has changed it in the mean time
				record = operations.get(operation);
			} else {
				ModelExtractionService.INSTANCE.newOperationRecord(record);
			}
		}
		return record;
	}
	
	public Collection<OperationRecord> getOperations() {
		return operations.values();
	}
	
	@Override
	public boolean isFiltered() {
		return false;
	}
	
	@Override
	public void accept(TraceVisitor visitor, InvocationGroupRecord source) {
//		System.out.println("Visit interface " + getInterfaceName());
		if (source != null) {
			visitor.visitRequiredInterface(source, this);
		} else {
			visitor.visitInterface(this);
		}
		
		for (OperationRecord orec : operations.values()) {
			orec.accept(visitor, source);
		}
	}

}
