package tools.descartes.prisma.agent.wildfly.interceptors;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.util.Deque;

import org.apache.commons.lang.mutable.MutableLong;
import org.jboss.ejb.client.EJBClientInterceptor;
import org.jboss.ejb.client.EJBClientInvocationContext;
import org.jboss.logging.Logger;

import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InterfaceRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationGroupRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.ThreadCallStack;

public class EJBOutgoingCallInterceptor implements EJBClientInterceptor {
	
	private static final Logger log = Logger.getLogger(EJBOutgoingCallInterceptor.class);
	
	private final ThreadMXBean threadBean;
	
	public EJBOutgoingCallInterceptor() {
		threadBean = ManagementFactory.getThreadMXBean();
		if (threadBean.isCurrentThreadCpuTimeSupported()) {
			threadBean.setThreadCpuTimeEnabled(true);
		}
	}
	
	private static final ThreadLocal<MutableLong> invocationTime = new ThreadLocal<MutableLong>() {
		protected MutableLong initialValue() {
			return new MutableLong();
		}
	};
	
	private static final ThreadLocal<MutableLong> startCpuTime = new ThreadLocal<MutableLong>() {
		protected MutableLong initialValue() {
			return new MutableLong();
		}
	};

	@Override
	public void handleInvocation(EJBClientInvocationContext ctx) throws Exception {
		Deque<InvocationGroupRecord> threadCallStack = ThreadCallStack.get();
		InvocationGroupRecord currentOperation = threadCallStack.peek();
		if (currentOperation != null) {
			String component = ctx.getLocator().getAppName() + "/" + ctx.getLocator().getModuleName() + "/" + ctx.getLocator().getBeanName();
			String interf = ctx.getInvokedMethod().getDeclaringClass().getName();
			String operation = ctx.getInvokedMethod().getName();
			
			ComponentRecord crec = currentOperation.requireComponent(component);
			InterfaceRecord irec = crec.requireInterface(interf);
			OperationRecord orec = irec.requireOperation(operation);
			
			InvocationGroupRecord grec = getInvocationGroup(orec, ctx);
			
			threadCallStack.push(grec);
			
			invocationTime.get().setValue(System.nanoTime());
//			startCpuTime.get().setValue(threadBean.getCurrentThreadCpuTime());
		}
		ctx.sendRequest();
	}

	@Override
	public Object handleInvocationResult(EJBClientInvocationContext ctx) throws Exception {
		try {
			return ctx.getResult();
		} finally {
			long endTime = System.nanoTime();
//			long endCpuTime = threadBean.getCurrentThreadCpuTime();
			Deque<InvocationGroupRecord> threadCallStack = ThreadCallStack.get();
			if (!threadCallStack.isEmpty()) {
				long executionTime = endTime - invocationTime.get().longValue();
//				long cpuTime = endCpuTime - startCpuTime.get().longValue();
				InvocationGroupRecord grec = threadCallStack.pop();
				grec.addInvocation(executionTime, 0, 0);		
					
				if (log.isTraceEnabled()) {
					InvocationGroupRecord caller = threadCallStack.peek();
					log.trace("Outgoing invocation (" + caller.getOperation().getInterface().getInterfaceName() + ", " + caller.getOperation().getOperationName() 
							+ ") -> (" + grec.getOperation().getInterface().getInterfaceName() + ", " + grec.getOperation().getOperationName() + ", " + executionTime + "ms)");				
				}
			}
		}		
	}
	
	private InvocationGroupRecord getInvocationGroup(OperationRecord orec, EJBClientInvocationContext ctx) {
		int[] names = orec.getObservedParameterIndeces();
		InvocationGroupRecord grec = null;
		if (names.length > 0) {
			int idx = orec.characterizeValue(0, ctx.getParameters()[names[0]]);
			grec = orec.requireInvocationGroup(idx);
		} else {
			grec = orec.requireInvocationGroup(0);
		}
		return grec;
	}

}
