package tools.descartes.prisma.agent.wildfly.interceptors;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.util.Deque;

import org.apache.commons.lang.mutable.MutableLong;
import org.jboss.as.ejb3.component.EJBComponent;
import org.jboss.invocation.InterceptorContext;
import org.jboss.logging.Logger;

import io.undertow.server.ExchangeCompletionListener;
import io.undertow.server.HttpHandler;
import io.undertow.server.HttpServerExchange;
import io.undertow.servlet.handlers.ServletRequestContext;
import tools.descartes.prisma.agent.wildfly.Constants;
import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InterfaceRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationGroupRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.ThreadCallStack;
import tools.descartes.prisma.agent.wildfly.service.MonitoringService;

public class IncomingServletCallHandler implements HttpHandler {
	
	private static final Logger log = Logger.getLogger(IncomingServletCallHandler.class);
	
	private final ThreadMXBean threadBean;
	
	private HttpHandler next;
	
	public IncomingServletCallHandler(HttpHandler next) {
		this.next = next;
		threadBean = ManagementFactory.getThreadMXBean();
		if (threadBean.isCurrentThreadCpuTimeSupported()) {
			threadBean.setThreadCpuTimeEnabled(true);
		}
	}
	
	@Override
	public void handleRequest(HttpServerExchange req) throws Exception {
		if (req.isInIoThread()) {
			req.dispatch(this);
			return;
	    }
		
		Deque<InvocationGroupRecord> threadCallStack = ThreadCallStack.get();
		String component = req.getRequestPath();
		String interf = "http:" + component;
		Deque<String> actionParamter = req.getQueryParameters().get("action");
		if (actionParamter != null) {
			String operation = actionParamter.getFirst();		
			ComponentRecord crec = MonitoringService.INSTANCE.getTraceRoot().requireComponent(component);
			if (crec.getComponentClass() == null) {
				crec.setComponentClass(component);
			}
			InterfaceRecord irec = crec.requireInterface(interf);
			OperationRecord orec = irec.requireOperation(operation);
			InvocationGroupRecord grec = getInvocationGroup(orec, req);
			threadCallStack.push(grec);			
		}
		next.handleRequest(req);

		long endTime = System.nanoTime();
		if (!threadCallStack.isEmpty()) {
			long startTime = req.getRequestStartTime();
			InvocationGroupRecord grec = threadCallStack.pop();
			long executionTime = endTime - startTime;
			if (log.isTraceEnabled()) {
				log.info("Incoming invocation (" + grec.getOperation().getInterface().getComponent().getComponentName() + ", " + grec.getOperation().getOperationName()
						+ ", " + executionTime + "ms)");
			}
//			grec.addInvocation(executionTime, 0, 0);
		}
	}

	
	private InvocationGroupRecord getInvocationGroup(OperationRecord orec, HttpServerExchange req) {
		String[] names = orec.getObservedParameterNames();
		InvocationGroupRecord grec = null;
		if (names.length > 0) {
			Deque<String> params = req.getQueryParameters().get(names[0]);
			if (params != null && !params.isEmpty()) {
				String value = params.getFirst();
				int idx = orec.characterizeValue(0, (value != null && value.isEmpty()) ? null : value);
				grec = orec.requireInvocationGroup(idx);
			} else {
				int idx = orec.characterizeValue(0, null);
				grec = orec.requireInvocationGroup(idx);
			}			
		} else {
			grec = orec.requireInvocationGroup(0);
		}
		return grec;
	}

}
